clear
close all

%% generate data
% generates n-dimensional 2-class data with normal distributions
dim = 3; %dimension of the samples
numSamples = 5000; %total number of samples
ratio = 5/1; %pos/neg samples ratio
mean1 = [1 2.5 2]; %mean of the first gaussian (positive samples)
mean2 = [0 0 2]; %mean of the second gaussian (negative samples)
variance = 1; %the variance is equal for the both gaussians

numPosSamples = round(numSamples*(ratio/(1+ratio)));% number of positive samples
numNegSamples = round(numSamples/(ratio+1));% number of negative samples

%data is a structure with samples and class assignment
samples = repmat(mean1,numPosSamples,1)+variance*randn(numPosSamples,dim);
data.samples = [samples; repmat(mean2,numNegSamples,1)+variance*randn(numNegSamples,dim)];
data.class = [ones(numPosSamples,1); zeros(numNegSamples,1)];

%plot generated data
plot3(data.samples(1:numPosSamples,1),data.samples(1:numPosSamples,2),data.samples(1:numPosSamples,3),'r.');
hold on
plot3(data.samples(numPosSamples+1:end,1),data.samples(numPosSamples+1:end,2),data.samples(numPosSamples+1:end,3),'b.');
legend('Positive Data', 'Negative Data');
grid on

%select the training data
trainSize = 500;
perm = randperm(numSamples); %permutate the data before training data selection
trainData.samples = data.samples(perm(1:trainSize),:);
trainData.class = data.class(perm(1:trainSize),:);

testData.samples = data.samples(perm(trainSize+1:end),:);
testData.class = data.class(perm(trainSize+1:end),:);

%% Classifiers learning
nb = NaiveBayes.fit(trainData.samples,trainData.class);
tree = classregtree(trainData.samples,trainData.class);
p = profile('info');
%% ROC construction and AUC computing
c1output = sumClass(testData.samples);
c2output = nb.posterior(testData.samples);
c2output = c2output(:,2);

ROC1 = ROCconstr(c1output,testData.class); % ROC construction
AUC1 = AUCcomp(ROC1); %AUC computing
ROC2 = ROCconstr(c2output,testData.class);
AUC2 = AUCcomp(ROC2);

%ROC plotting
figure
title('ROC curve');
plot([0 1],[0 1]);
hold on
plot(1-ROC1.specificity,ROC1.sensitivity,'g');
plot(1-ROC2.specificity,ROC2.sensitivity,'k');
leg1 = ('Random classifier, AUC = 0.5');
leg2 = sprintf('ROC of the %s, AUC = %0.3f','Sum classifier',AUC1);
leg3 = sprintf('ROC of the %s, AUC = %0.3f','Naive Bayes classifier',AUC2);
legend(leg1,leg2,leg3);
xlabel('FPR (1-Specificity)');
ylabel('TPR (Sensitivity)');


%% ROC convex hull
TPR = []; FPR = [];
bayesEst = nb.predict(testData.samples);
[TPR(end+1) FPR(end+1)] = computeTPRFPR(bayesEst,testData.class);
names{1} = 'NaiveBayes';

treeEst = tree.eval(testData.samples);
[TPR(end+1) FPR(end+1)] = computeTPRFPR(treeEst,testData.class);
names{2} = 'ClassificationTree';

knnEst = knn(trainData.samples,trainData.class,testData.samples);
[TPR(end+1) FPR(end+1)] = computeTPRFPR(knnEst,testData.class);
names{3} = 'k-nearest neighbours';

mahalEst = mahalClass(trainData.samples,trainData.class,testData.samples);
[TPR(end+1) FPR(end+1)] = computeTPRFPR(mahalEst,testData.class);
names{4} = 'Mahalanobis distance';

%convex hull
currPoint = [0 0];
curve = currPoint;
TPR(end+1) = 1; FPR(end+1) = 1;
while currPoint~=[1 1]
    subtr = [TPR' FPR']-repmat(currPoint,length(TPR),1);
    adeptInd = find(subtr(:,1)>=0 & subtr(:,1)>=0); 
    subtr1 = subtr(adeptInd,:);
    [val ind] = max(subtr1(:,1)./subtr1(:,2));
    currPoint = [TPR(adeptInd(ind)) FPR(adeptInd(ind))];
    curve = [curve; currPoint];
end
TPR(end) = []; FPR(end) = [];
clear subtr subtr1 val ind adeptInd currPoint

%plotting covnex hull
figure
title('ROC convex hull')
plot([0 1],[0 1]);
hold on
plot(FPR,TPR,'r.');
xlim([0 1]);
ylim([0 1]);
gname(names);
plot(curve(:,2),curve(:,1),'g:')

clear curve 

%% ROC from crossvalidation
n = 5;
foldSize = trainSize/n;
for i = 1:n    
    mask(ceil(foldSize*(i-1))+1:round(foldSize*i)) = i;
end
allOut = [];
allTestClass = [];
for i = 1:n
    cvTest.samples = trainData.samples(mask==i,:);
    cvTest.class = trainData.class(mask==i);
    cvTrain.samples = trainData.samples(mask~=i,:);
    cvTrain.class = trainData.class(mask~=i);
    cvBayes = NaiveBayes.fit(cvTrain.samples,cvTrain.class);
    output = cvBayes.posterior(cvTest.samples);
    output = output(:,2);
    allOut = [allOut; output];
    allTestClass = [allTestClass; cvTest.class];
    ROC{i} = ROCconstr(output,cvTest.class);
end

cvROC = ROCconstr(allOut,allTestClass);
cvPlot = ROCplot(cvROC);
legend('Mean ROC')
set(findobj(gca,'Type','line'),'Color','r')
for i = 1:n
    cvPlot = ROCplot(ROC{i},cvPlot);
end


